// 
//
// IwGame - Cross Platform Multi-purpose Game Engine using the Marmalade SDK
//
// Developed by Matthew Hopwood of Pocketeers Limited - www.pocketeers.co.uk
//
// For updates, tutorials and more details check out my blog at www.drmop.com
//
// This code is provided free of charge and without any warranty whatsoever. The only restriction to its usage is that this header must remain intact and visible in all IwGame engine files.
// If you use this engine in your product, whilst it is not mandatory, a mention of the IwGame engine would be nice or a quick email to let us know where it is being used.
//
//

#include "IwGameAnim.h"

//
//
// CIwGameAnim implementation
//
//
void CIwGameAnim::setLooped(int num_loops)
{
	// Mark as a looping animation and set the number of loops to play
	Looping = true;
	MaxLoops = num_loops;
	NumLoops = 0;
}

void CIwGameAnim::setCurrentFrame(float frame)
{
	// Set the current frame
	if ((int)CurrentFrame != (int)frame)
		FrameHasChanged = true;
	CurrentFrame = frame;
}

void CIwGameAnim::Restart()
{
	// Restart the animation from its start
	PlayDelay.Reset();
	CurrentFrame = 0;
	NumLoops = 0;
	Playing = true;
}

bool CIwGameAnim::Update(float dt)
{
	// Exit if stopped
	if (!Playing)
		return false;

	// Check to see if play delay is active, if so then wait until the play delay time has passed before starting the animation
	if (PlayDelay.hasStarted())
	{
		if (!PlayDelay.HasTimedOut())
			return false;
		else
		{
			// Let the user know that the animation has started
			if (StartedCallback != NULL)
				StartedCallback((void*)this, NULL);
		}
	}
	// Get the integer portion of the current frame
	int cf = (int)CurrentFrame;

	// Update the current frame
	CurrentFrame += PlaybackSpeed * dt;

	// if current frame has gone passed end or start of animation then wrap animation or end it
	if (CurrentFrame >= FrameCount || CurrentFrame < 0)
	{
		if (Looping)
		{
			// Count number of times the animation has been played
			NumLoops++;
			if (MaxLoops < 0)	// Repeat animation forever
			{
				// Loop the animation
				if (CurrentFrame < 0)
					CurrentFrame += FrameCount;
				else
					CurrentFrame -= FrameCount;

				// Notify use that the animation has looped
				if (LoopedCallback != NULL)
					LoopedCallback((void*)this, NULL);
			}
			else
			{
				if (NumLoops < MaxLoops)
				{
					// Loop the animation
					if (CurrentFrame < 0)
						CurrentFrame += FrameCount;
					else
						CurrentFrame -= FrameCount;

					// Notify use that the animation has looped
					if (LoopedCallback != NULL)
						LoopedCallback((void*)this, NULL);
				}
				else
				{
					// Animation has ran out of loops so inform the user and stop the animation
					if (StoppedCallback != NULL)
						StoppedCallback((void*)this, NULL);
					Stop();
				}
			}
		}
		else
		{
			// Animation has finished so inform the user and stop the animation
			if (StoppedCallback != NULL)
				StoppedCallback((void*)this, NULL);
			Stop();
		}
	}

	// If animation changed frames then mark animation as changed (user can pol this to check when to update something based on the curretn frame changing)
	if (cf != (int)CurrentFrame)
		FrameHasChanged = true;
	else
		FrameHasChanged = false;

	return true;
}

//
//
// CImageFrameAnimation implementation
//
//

//
//
// CIWGameAnimFloat implementation
//
//

float CIWGameAnimFloat::getCurrentFrameData()
{
	// Calculate the current frames value
	if (CurrentFrame >= 0)
	{
		float* frame = Frames + (int)CurrentFrame;
		float* next_frame = Frames + (int)CurrentFrame + 1;
		return  *frame + (*next_frame - *frame) * (CurrentFrame - (int)CurrentFrame);
	}

	return *Frames;
}

//
//
// CIwGameAnimFVec2  implementation
//
//
bool CIwGameAnimFVec2 ::Update(float dt)
{
	// Update animation system
	bool ret = CIwGameAnim::Update(dt);

	// Calculate current frame value
	if (CurrentFrame >= 0)
	{
		CIwFVec2* frame = Frames + (int)CurrentFrame;
		CIwFVec2* next_frame = Frames + (int)CurrentFrame + 1;
		CurrentData.x = frame->x + (next_frame->x - frame->x) * (CurrentFrame - (int)CurrentFrame);
		CurrentData.y = frame->y + (next_frame->y - frame->y) * (CurrentFrame - (int)CurrentFrame);
	}

	return ret;
}

//
//
// CIwGameAnimManager implementation
//
// 
CIwGameAnimManager::~CIwGameAnimManager()
{
	for (Iterator it = Animations.begin(); it != Animations.end(); ++it)
		delete *it;
	Animations.clear();
}

void CIwGameAnimManager::setCurrentAnimation(int animation_index)
{
	if (CurrentAnimation >= 0)
		Animations.element_at(animation_index)->Stop();
	Animations.element_at(animation_index)->Start();
	CurrentAnimation = animation_index;
}

CIwGameAnim* CIwGameAnimManager::getCurrentAnimation()
{
	if (CurrentAnimation >= 0)
		return Animations.element_at(CurrentAnimation);

	return NULL;
}

void CIwGameAnimManager::Update(float dt)
{
	if (UpdateAll)
	{
		// Mode 1 - Update all of the contained animations
		for (Iterator it = Animations.begin(); it != Animations.end(); ++it)
			(*it)->Update(dt);
	}
	else
	{
		// Mode 2 - Update only the current animation
		if (CurrentAnimation >= 0)
			Animations.element_at(CurrentAnimation)->Update(dt);
	}
}

//
//
// CIwGameAnimFrameManager implementation
//
// 
CIwGameAnimFrameManager::~CIwGameAnimFrameManager()
{
	// Clear up float animation frame sets
	for (CIwList<float*>::iterator it = FloatFrames.begin(); it != FloatFrames.end(); ++it)
		delete [] *it;
	FloatFrames.clear();

	// Clear up FVec2 animation frame sets
	for (CIwList<CIwFVec2*>::iterator it = FVec2Frames.begin(); it != FVec2Frames.end(); ++it)
		delete [] *it;
	FVec2Frames.clear();

	// Clear up image animation frame sets
	for (CIwList<CIwGameAnimImageFrame*>::iterator it = ImageFrames.begin(); it != ImageFrames.end(); ++it)
		delete [] *it;
	ImageFrames.clear();
}


float* CIwGameAnimFrameManager::allocFloatFrames(int count)
{
	float* data = new float[count];
	FloatFrames.push_back(data);

	return data;
}

CIwFVec2* CIwGameAnimFrameManager::allocFVec2Frames(int count)
{
	CIwFVec2* data = new CIwFVec2[count]();
	FVec2Frames.push_back(data);

	return data;
}

CIwGameAnimImageFrame* CIwGameAnimFrameManager::allocImageFrames(int count)
{
	CIwGameAnimImageFrame* data = new CIwGameAnimImageFrame[count]();
	ImageFrames.push_back(data);

	return data;
}

//
// allocImageFrames(count, frame_w, frame_h, start_x, start_y, pitch_x, pitch_y, image_width)
// - Helper method to automate generation of a group of frames located on a sprite atlas. Calling this method will generate the specified frames at the specified size
// - Paremeters:
// -- count - Total frames to generate
// -- frame_w, frame_h - Width and height of each animation frame
// -- start_x, start_y - Upper left hand corner of first animation frame
// -- pitch_x, pitch)y - The horizontal and vertical spacing between frames
// -- image_width		- Width of the sprite atlas
// - 
// - Returns: An array of image animation frames
//
CIwGameAnimImageFrame* CIwGameAnimFrameManager::allocImageFrames(int count, int frame_w, int frame_h, int start_x, int start_y, int pitch_x, int pitch_y, int image_width)
{
	int						x = start_x; 
	int						y = start_y;
	CIwGameAnimImageFrame*	data = new CIwGameAnimImageFrame[count]();
	CIwGameAnimImageFrame*	frames = data;
	ImageFrames.push_back(data);

	while (count != 0)
	{
		// Set frames data
		frames->x = x;
		frames->y = y;
		frames->w = frame_w;
		frames->h = frame_h;

		// Set to next frame horizontally
		x += pitch_x;
		if (x >= image_width)
		{
			// Step down to next line of frames
			y += pitch_y;
			x = start_x;
		}

		frames++;
		count--;
	}

	return data;
}


