// 
//
// IwGame - Cross Platform Multi-purpose Game Engine using the Marmalade SDK
//
// Developed by Matthew Hopwood of Pocketeers Limited - www.pocketeers.co.uk
//
// For updates, tutorials and more details check out my blog at www.drmop.com
//
// This code is provided free of charge and without any warranty whatsoever. The only restriction to its usage is that this header must remain intact and visible in all IwGame engine files.
// If you use this engine in your product, whilst it is not mandatory, a mention of the IwGame engine would be nice or a quick email to let us know where it is being used.
//
//

#include "IwGameScene.h"
#include "IwGameActor.h"
#include "IwGameCamera.h"

//
//
// CIwGameScene implementation
//
//

CIwGameScene::~CIwGameScene()
{
	if (Collidables != NULL)
	{
		delete Collidables;
		Collidables = NULL;
	}
	clearActors();
	if (SpriteManager != NULL)
		delete SpriteManager;
	if (AnimFrameManager != NULL)
		delete AnimFrameManager;
	if (ImageManager != NULL)
		delete ImageManager;
	if (Camera != NULL)
		delete Camera;
}

void CIwGameScene::setVirtualTransform(int required_width, int required_height, float angle, bool fix_aspect, bool lock_width)
{
	float scale_x, scale_y;
	int screen_width = IwGxGetScreenWidth();
	int screen_height = IwGxGetScreenHeight();
	
	ScreenSize.x = screen_width;
	ScreenSize.y = screen_height;
	VirtualSize.x = required_width;
	VirtualSize.y = required_height;
	if (fix_aspect)
	{
		// A fixed aspect ensures that the aspect ratio of the scene fits the native screen aspect ratio
		if (lock_width)
			scale_x = scale_y = (float)screen_width / required_width;
		else
			scale_x = scale_y = (float)screen_height / required_height;
	}
	else
	{
		// Simply scale to fit native screen resolution
		scale_x = (float)screen_width / required_width;
		scale_y = (float)screen_height / required_height;
	}

	// Build rotation / scale tranform
	CIwMat2D scale;
	scale.SetIdentity();
	scale.m[0][0] = (iwfixed)(scale_x * IW_GEOM_ONE);
	scale.m[1][1] = (iwfixed)(scale_y * IW_GEOM_ONE);
	VirtualTransform.SetRot((iwangle)((angle * IW_ANGLE_2PI) / 360.0f));
	VirtualTransform.PostMult(scale);

	// Cantre scene in middle of screen
	VirtualTransform.t.x = screen_width  / 2;
	VirtualTransform.t.y = screen_height  / 2;

	// If camera attacjhed then force transform to become dirty so that the scene transform gets rebuilt
	if (Camera != NULL)
	{
		Camera->forceTransformDirty();
	}
}

int	 CIwGameScene::Init(int max_collidables, int max_layers)
{
	ClippingArea.x = 0;
	ClippingArea.y = 0;
	ClippingArea.w = 0;
	ClippingArea.h = 0;

	// Allocate collidables list
	Collidables = new CIwGameActor*[max_collidables];
	MaxCollidables = max_collidables;
	MaxLayers = max_layers;

	// Allocate managers
	SpriteManager = new CIwGameSpriteManager();
	SpriteManager->Init(max_layers);
	AnimFrameManager = new CIwGameAnimFrameManager();
	ImageManager = new CIwGameImageManager();

	return 1;
}

void CIwGameScene::addActor(CIwGameActor *actor)
{
	// Add an actor to the scene
	Actors.push_back(actor);
	actor->setScene(this);
}

void CIwGameScene::removeActor(CIwGameActor* actor)
{
	// Renove named actor from the scene
	for (_Iterator it = Actors.begin(); it != Actors.end(); ++it)
	{
		if (*it == actor)
		{
			// Do not delete managed actors as the pooled object manager will be handling this
			if (!(*it)->isManaged())
				delete *it;
			else
			{
				// Instead we disable them and mark them as unused so they can be reallocated
				(*it)->setUsed(false);
				(*it)->setActive(false);
			}
			Actors.erase(it);
			break;
		}
	}
}

void CIwGameScene::removeActor(unsigned int name_hash)
{
	// Renove named actor from the scene
	for (_Iterator it = Actors.begin(); it != Actors.end(); ++it)
	{
		if (name_hash == (*it)->getNameHash())
		{
			// Do not delete managed actors as the pooled object manager will be handling this
			if (!(*it)->isManaged())
				delete *it;
			else
			{
				// Instead we disable them and mark them as unused so they can be reallocated
				(*it)->setUsed(false);
				(*it)->setActive(false);
			}
			Actors.erase(it);
			break;
		}
	}
}

CIwGameActor* CIwGameScene::findActor(unsigned int name_hash)
{
	// Find and eturn named actor
	for (_Iterator it = Actors.begin(); it != Actors.end(); ++it)
	{
		if (name_hash == (*it)->getNameHash())
			return *it;
	}

	return NULL;
}

CIwGameActor* CIwGameScene::findActor(int type)
{
	// Find and return first actor of specified type
	for (_Iterator it = Actors.begin(); it != Actors.end(); ++it)
	{
		if (type == (*it)->getType())
			return *it;
	}
	return NULL;
}

void CIwGameScene::clearActors()
{
	// Remove and delete all actors from the scsne (only delete if actors are not managed)
	for (_Iterator it = Actors.begin(); it != Actors.end(); ++it)
	{
		if (!(*it)->isManaged())
			delete *it;
	}
	Actors.clear();
}

bool CIwGameScene::addCollideable(CIwGameActor* actor)
{
	if (NextFreeCollidable < MaxCollidables)
		Collidables[NextFreeCollidable++] = actor;
	else
		return false;

	return true;
}

void CIwGameScene::Update(float dt)
{
	if (Camera != NULL)
	{
		// I need to point out here that we use the concept of dirty transforms to cut down processing when  the camera is not moving
		// If the cameras transform is not dirty (the camera has not moved) then there is no need to force the sprite manager to make 
		// all of its child sprites to rebuild their transforms
		if (Camera->isTransformDirty())
		{
			// Update camera transform
			Camera->Update();

			// Combine the virtual canvas and camera transforms
			Transform = Camera->getTransform();
			Transform.PostMult(VirtualTransform);

			// Write camera transform to the sprite manager
			if (SpriteManager != NULL)
			{
				SpriteManager->setTransform(Transform);
			}
		}
	}

	// Process the actors
	CIwList<CIwGameActor*> removals;
	for (_Iterator it = begin(); it != end(); ++it)
	{
		if (!(*it)->Update(dt))
			removals.push_back(*it);
	}

	// Resolve actor collisions
	for (_Iterator it = begin(); it != end(); ++it)
		(*it)->ResolveCollisions();

	// Remove any deleted actors
	for (_Iterator it = removals.begin(); it != removals.end(); ++it)
		removeActor(*it);

	// Reset collidables list
	NextFreeCollidable = 0;
}

void CIwGameScene::Draw()
{
	// Transfor actor visual data to sprite manager sprites
	for (_Iterator it = begin(); it != end(); ++it)
		(*it)->UpdateVisual();

	if (ClippingArea.w != 0)
	{
		// Set clipping rectangle (clipping coords are specified in virtual canvas coordinates so we need to transform them to the screen)
		CIwSVec2 cliprect[2];
		cliprect[0].x = ClippingArea.x;
		cliprect[1].x = ClippingArea.x + ClippingArea.w;
		cliprect[0].y = ClippingArea.y;
		cliprect[1].y = ClippingArea.y + ClippingArea.h;
		for (int t = 0; t < 2; t++)
			cliprect[t] = VirtualTransform.TransformVec(cliprect[t]);
		IwGxClearScissorScreenSpace();
		IwGxSetScissorScreenSpace(cliprect[0].x, cliprect[0].y, cliprect[1].x - cliprect[0].x, cliprect[1].y - cliprect[0].y);
	}

	// Draw the sprite manager sprites
	SpriteManager->Draw();

}

CIwFVec2 CIwGameScene::ScreenToVirtual(float screen_x, float screen_y)
{
	screen_x -= ScreenSize.x / 2;
	screen_y -= ScreenSize.y / 2;
	return CIwFVec2((screen_x * IW_GEOM_ONE) / Transform.m[0][0], (screen_y * IW_GEOM_ONE) / Transform.m[1][1]);
}

CIwVec2 CIwGameScene::VirtualToScreen(float virtual_x, float virtual_y)
{
	return CIwVec2((virtual_x * Transform.m[0][0]) / IW_GEOM_ONE + ScreenSize.x / 2, (virtual_y * Transform.m[1][1]) / IW_GEOM_ONE + ScreenSize.y / 2);
}



